import ALIASES from './SyrusLang.d/aliases.js'
import { COMMANDS, PERSISTENT_COMMANDS } from './SyrusLang.d/commands.js'
import { getDegree, getDistance, getTime, getAsCSVPair } from './SyrusLang.d/utils.js'

/**
 * list of arguments passed on the syruslang line that must be converted to units
 * ej: min_duration=10min, speed_threshold=20kph
*/
const TO_UNITS = [
	"min_duration",
	"odometer",
	"radius",
	"speed_threshold",
	"rpm_threshold",
	"begin_idle_time",
	"ignition_time",
	"idle_time",
	"over_speed",
	"over_rpm",
	"hard_brakes",
	"harsh_fwd_acceleration"
];

export class SyrusLang {
	constructor() {}

	/**
	 * replace in syruslang Line aliases for equivalent
	 * @param {*} command syruslang line  config
	 */
	checkAliases(command) {
		for (const alias in ALIASES) {
			if (command.startsWith(alias)) return command.replace(alias, ALIASES[alias]);
		}
		return command;
	}

	/**
	 * function used to convert syruslang define geofence in geojson
	 * @param {*} latlngsString the latlngs part of the syruslang
	 * @param {*} radius the radius part of the syruslang
	 * @param {*} type the type of the geofence could be auto,Polygon or Circular
	 */
	toGeoJSON(latlngsString, radius, type = "auto") {
		let regexCircular = /^(\s?\-?\d{1,99}\.?\d{1,99}\,\-?\d{1,99}\.?\d{1,99}\s?){1,1}$/g;
		if (regexCircular.test(latlngsString)) {
			return {
				type: "Feature",
				geometry: {
					coordinates: latlngsString.split(",").map(l => Number(l)),
					radius: radius || 50,
					type: "Circle"
				},
				properties: {
					radius: radius
				}
			};
		}
		var regex = /^(\s?\-?\d{1,99}\.?\d{1,99}\,\-?\d{1,99}\.?\d{1,99}\,?){2,}(\,?\-?\d{1,99}\.?\d{1,99}\,\-?\d{1,99}\.?\d{1,99}){0,}\s?$/g;
		if (!regex.test(latlngsString)) {
			throw "invalid latlngs " + latlngsString;
		}
		var coords = latlngsString
			.trim()
			.split(" ")
			.map(latlng => {
				return latlng.split(",").map(l => Number(l));
			});
		var type = radius > 0 ? "LineString" : "Polygon";
		if (type == "Polygon") {
			var first = coords[0];
			var last = coords[coords.length - 1];
			if (first[0] !== last[0] || first[1] !== last[1]) {
				coords.push([first[0], first[1]]);
			}
		}
		return {
			type: "Feature",
			geometry: {
				coordinates: type == "LineString" ? coords : [[coords]],
				radius: radius,
				type: type
			},
			properties: {
				radius: radius
			}
		};
	}

	/**
	 * Entry point of this class where all the syruslang is received
	 * @param {*} text accepts Syruslang lines, could be a single command or the whole config
	 * @returns array of SyrusLang Json actions, ignores invalid commands
	 */
	process(text) {
		var lines = text.split(/\r?\n/);
		lines = lines.filter(c => {
			if (c.startsWith("#") || c == "") {
				return false;
			}
			return true;
		});
		var commands = [];
		lines.forEach(line => {
			var command = this._process(line.trim());
			if (command) {
				command.raw = line;
				commands.push(command);
			}
		});
		return commands.filter(v => !!v);
	}

	/**
	 * !Important Single line process of syruslang parser
	 * !All the important logic starts here
	 * @param {*} _line syrusLang line command
	 * @returns syruslang action object
	 */
	_process(_line) {
		let line = this.checkAliases(_line); // onvert alias to syruslang
		let command;
		// Checkign is a valid command
		for (const cmd of COMMANDS) {
			if (line.startsWith(cmd)) {
				command = cmd;
				break;
			}
		}
		if (!command) {
			console.warn("not supported: ", line);
			return;
		}
		// Getting all parameters from syruslang
		let args = this.getArguments(line);
		args.command = command;
		args.persistent = PERSISTENT_COMMANDS.includes(command); // if persisteng flag is set to append to the appends.syrus.conf

		// Specials
		// Has unix commands
		if (args.command == "exec") {
			return {
				command: "exec",
				argument: line.replace("exec", "")
			};
		}
		// Has syruslang command
		if (args.command == "define action") {
			let name = line.split(" ")[2];
			let signals = this.getArg(line, "trigger", false)
			if (signals) signals = signals.split(',')
			return {
				name,
				signals,
				command: "define action",
				persistent: true,
				rate: args.rate ? this.getRater(args.rate) : null,
				action: line
					.replace(`define action ${name}`, "")
					.replace(`trigger=${this.getArg(line, "trigger", false)}`, "")
					.replace(`rate=${this.getArg(line, "rate", false)}`, "")
					.trim()
			};
		}
		// design to delete the whole config
		if (args.command == "FLUSH ALL") {
			return {
				command: "FLUSH ALL"
			};
		}

		// Depending of the command the different argument must be processed
		if (args.command == "define destination") {
			let [, , name, protocol, url] = line
				.split(" ")
				.map(arg => arg.trim())
				.filter(m => m);
			args.name = name;
			args.protocolId = args.protocol;
			args.protocol = protocol;
			args.persistent = true;
			if (args.channelRx && !args.subscribe) {
				args.subscribe = args.channelRx;
				console.warn("channelRx is now deprecated you should use subscribe instead [mqtt destination]");
			}
			if (args.channelTx && !args.publish) {
				console.warn("channelTx is now deprecated you should use publish instead [mqtt destination]");
				args.publish = args.channelTx;
			}
			args.channel = url.split("://")[0];
			args.host = url.split("://")[1].split(":")[0];
			args.port = url.split("://")[1].split(":")[1] || 5000;
			if (args.headers) args.headers = getAsCSVPair(args.headers);
			if (args.params) args.params = getAsCSVPair(args.params);
		} else if (
			args.command == "define tracking_resolution" ||
			args.command == "set tracking_resolution"
		) {
			let [, , name] = line.split(" ").map(arg => arg.trim());
			let heading = getDegree(line);
			let time = getTime(line);
			let distance = getDistance(line);
			args.name = name;
			args.time = this.toBasicUnit(time) || 0;
			args.heading = this.toBasicUnit(heading) || 0;
			args.distance = this.toBasicUnit(distance) || 0;
			args.persistent = args.command.includes("define ");
		} else if (args.command == "define signal") {
			let name = line.split(" ")[2];
			let [variable, operator, comparator] = line
				.replace(`define signal ${name}`, "")
				.replace(`min_duration=${this.getArg(line, "min_duration", false)}`, "")
				.trim()
				.split(" ")
				.map(arg => arg.trim());
			args.name = name;
			args.persistent = true;
			args.variable = variable;
			args.operator = operator;
			args.comparator = comparator;
		} else if (args.command == "define group") {
			args.persistent = true;
			args.name = line.split(" ")[2];
			args.meta = line.split(" ")[3];
			args.fieldset = args.fieldset || "default";
			args.rate = args.rate ? this.getRater(args.rate) : null;
			args.destinations = this.getArg(line, "destinations", false)
			if (args.destinations) args.destinations = args.destinations.split(",")
		} else if (args.command == "define fieldset") {
			args.name = line.split(" ")[2];
			args.fields = this.fieldset(line);
			args.persistent = true;
		} else if (args.command == "define event") {
			let [, , name] = line.split(" ").map(arg => arg.trim());
			args.name = name;
			args.code = args.code || "01";
			args.label = args.label || "trckpnt";
			args.group = args.group || "default";
			args.fieldset = args.fieldset || "default";
			args.ack = args.ack;
			args.retain = `${args.retain}` == "true";
			args.destinations = this.getArg(line, "destinations", false)
			if (args.destinations) args.destinations = args.destinations.split(",")
			args.signals = this.getArg(line, "trigger", false);
			if (args.signals) args.signals = args.signals.split(",")
			args.rate = args.rate ? this.getRater(args.rate) : null;
			args.photo = this.getArg(line, "photo", false)
		} else if (args.command == "define geofence") {
			let [, , name] = line.split(" ", 3);
			args.name = name;
			args.group = this.getArg(line, "group", false)
			args.group = args.group ? args.groups.split(',') : []
			args.lngLats = line
				.replace(name, "")
				.replace("define geofence", "")
				.replace(`group=${args.group.join(",")}`, "")
				.replace(`radius=${this.getArg(line, "radius", false)}`, "")
				.trim();
		} else if (args.command == "define counters") {
			let [, , name] = line.split(" ", 3);
			args.name = name;
			args.forceSet = false;
		} else if (args.command == "define timer"){
			let [, , name] = line.split(" ", 3)
			args.name = name
			args.repeat = this.getArg(line, "repeat", false) || false
			args.enabled = this.getArg(line, "enabled", false) || false
			let duration = this.getArg(line, "duration", true)
			let delay = this.getArg(line, "delay", true)
			args.duration = !isNaN(parseInt(duration)) ? parseInt(duration) : 0
			args.delay = !isNaN(parseInt(delay)) ? parseInt(delay) : 0
			if (!args.duration){
				console.warn(`define timer ${name}: missing duration argument`)
			}
		} else if (args.command == "set namespace") {
			args.name = line.split(" ")[2];
		} else if (args.command == "set destinations") {
			args.groups = args.groups ? args.groups : args.group
			args.events = args.events ? args.events : args.event
			args.destinations = line
				.split(" ")
				.reverse()[0]
				.split(",");
		} else if (args.command == "set variable") {
			let [, , name, value] = line.split(" ", 4);
			if (value.endsWith('"')) value = value.substring(2, value.length - 2);
			args.name = name;
			args.value = value;
		} else if (args.command == "set counters") {
			let [, , name] = line.split(" ", 3);
			args.name = name;
			args.forceSet = true;
		} else if (args.command == "set @ecu_params_response") {
			let [, key, val] = line.split(" ", 3);
			args.key = key;
			args.val = val;
		} else if (args.command == "set out") {
			let [, io, state] = line.split(" ");
			args.io = io.replace("$io.out", "").replace("out", "");
			args.state = state;
		} else if (args.command == "set value") {
			let [, , name, value] = line.split(" ");
			args.variable_name = name;
			args.value = value;
		} else if (args.command == "start counters") {
			let [, , name] = line.split(" ", 3);
			args.name = name;
		} else if (args.command == "stop counters") {
			let [, , name] = line.split(" ", 3);
			args.name = name;
		} else if (args.command == "delete") {
			let [, type, name] = line.split(" ");
			args.name = name;
			args.type = type;
		} else if (args.command == "enable" || args.command == "start") {
			let [_, type, name] = line.split(" ");
			args.type = type;
			args.name = name;
			args.state = "enable";
		} else if (args.command == "disable" || args.command == "stop") {
			let [_, type, name] = line.split(" ");
			args.type = type;
			args.name = name;
			args.state = "disable";
		} else if (args.command == "get") {
			let [, type, name] = line.split(" ");
			args.type = type.replace("_value", "");
			args.variable_name = name;
			args.returnValue = type.endsWith("_value");
		} else if (args.command == "send event") {
			let [, , name] = line.split(" ").map(arg => arg.trim());
			args.name = name;
		} else if (args.command == "trigger action") {
			let [, , name] = line.split(" ").map(arg => arg.trim());
			args.name = name;
		} else if (args.command == "fire signal") {
			let [, , name] = line.split(" ").map(arg => arg.trim());
			args.signal_name = name;
		}
		return args;
	}

	/**
	 *  function to get single argument from a syruslang command fieldset="..."
	 * @param {*} line the line itself to extract the argument
	 * @param {*} variableName  the name of the argument
	 * @param {*} convert true if the argument must be convert to units like time,distance,rpm,speed, etc...
	 */
	getArg(line, variableName, convert = true) {
		let index = line.indexOf(`${variableName}=`);
		if (index == -1) return null
		let value = line.split(`${variableName}=`)[1]
		if (value.startsWith('"')) value = value.split('"')[1]
		else if (value.startsWith("'")) value = value.split("'")[1]
		else value = value.split(' ')[0]
		value = JSON.parse(`"${value}"`)
		if (convert) value = this.toBasicUnit(value);
		return value;
	}

	/**
	 * Spacial function to get the rater argument that it's composed in two parts: limit/time
	 * @param {*} val The values from the rater
	 */
	getRater(val) {
		let [count, time] = val.split("/");
		time = this.toBasicUnit(time);
		return {
			times: count,
			timeout: time
		};
	}

	/**
	 * Get all the arguments from a syruslang command
	 * @param {string} line the syruslang line command
	 * @returns An object with the values of the all arguments
	 */
	getArguments(line) {
		const regex = /\w{1,}=\S{1,}/gm;
		let m;
		let args = {};
		while ((m = regex.exec(line)) !== null) {
			// This is necessary to avoid infinite loops with zero-width matches
			if (m.index === regex.lastIndex) {
				regex.lastIndex++;
			}

			// The result can be accessed through the `m`-variable.
			m.forEach(match => {
				let [key] = match.split("=");
				let convert = TO_UNITS.includes(key);
				args[key] = this.getArg(line, key, convert);
			});
		}
		return args;
	}

	/**
	 * function to convert units in the units accepted by syrusjs example 1hr or 1min to ms that is expected for syrusjs
	 * @param {string} value to convert include the unit
	 */
	toBasicUnit(value) {
		if (!isNaN(parseInt(value))) {
			// Distance
			if (value.endsWith("km")) {
				value = parseInt(value) * 1000;
			} else if (value.endsWith("mts")) {
				value = parseInt(value) * 1;
			} else if (value.endsWith("mile")) {
				value = parseInt(value) * 1609;
			} else if (value.endsWith("ft")) {
				value = parseInt(parseInt(value) * 0.3048);
			}
			// Speed
			else if (value.endsWith("kph")) {
				value = parseInt(value) * 1;
			} else if (value.endsWith("mph")) {
				value = parseInt(value) * 1.60934;
			} else if (value.endsWith("m/s")) {
				value = parseInt(value) * 3.6;
			}
			// Time
			else if (value.endsWith("s") || value.endsWith("sec")) {
				value = parseInt(value) * 1000;
			} else if (value.endsWith("h") || value.endsWith("hr")) {
				value = parseInt(value) * 1000 * 3600;
			} else if (value.endsWith("m") || value.endsWith("min")) {
				value = parseInt(value) * 60 * 1000;
			}
			// Degree
			else if (value.endsWith("deg")) {
				value = parseInt(value);
			} else if (value.endsWith("rad")) {
				value = parseInt(parseInt(value) * (180 / Math.PI));
			}
			// Cycles
			else if (value.endsWith("hz")) {
				value = parseInt(value) * 60;
			} else if (value.endsWith("rpm")) {
				value = parseInt(value) * 1;
			}
			// default
			else {
				value = parseInt(value);
			}
		}
		if (isNaN(value)) return null
		return value
	}

	/**
	 * helper function to process the fieldset argument
	 * @param {*} line the syruslang line command
	 * @returns the fields injavascript object
	 */
	fieldset(line) {
		var _fields = line.substring(line.indexOf(`fields=`) + 7).split(",");
		var fields = {};
		for (const field of _fields) {
			var [key, value] = field.split(":");
			fields[key] = value ? value : key;
		}
		return fields;
	}
}
